/* Copyright (c) 2010 - 2018, Nordic Semiconductor ASA
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without modification,
 * are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice, this
 * list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form, except as embedded into a Nordic
 *    Semiconductor ASA integrated circuit in a product or a software update for
 *    such product, must reproduce the above copyright notice, this list of
 *    conditions and the following disclaimer in the documentation and/or other
 *    materials provided with the distribution.
 *
 * 3. Neither the name of Nordic Semiconductor ASA nor the names of its
 *    contributors may be used to endorse or promote products derived from this
 *    software without specific prior written permission.
 *
 * 4. This software, with or without modification, must only be used with a
 *    Nordic Semiconductor ASA integrated circuit.
 *
 * 5. Any software provided in binary form under this license must not be reverse
 *    engineered, decompiled, modified and/or disassembled.
*
 * THIS SOFTWARE IS PROVIDED BY NORDIC SEMICONDUCTOR ASA "AS IS" AND ANY EXPRESS
 * OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY, NONINFRINGEMENT, AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL NORDIC SEMICONDUCTOR ASA OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
 * GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <stdint.h>
#include <string.h>

/* HAL */
#include "boards.h"
#include "simple_hal.h"
#include "app_timer.h"

/* Core */
#include "nrf_mesh_config_core.h"
#include "nrf_mesh_gatt.h"
#include "nrf_mesh_configure.h"
#include "nrf_mesh.h"
#include "mesh_stack.h"
#include "device_state_manager.h"
#include "access_config.h"
#include "proxy.h"

/* Provisioning and configuration */
#include "mesh_provisionee.h"
#include "mesh_app_utils.h"

/* Models */
#include "generic_onoff_server.h"

/* Logging and RTT */
#include "log.h"
#include "rtt_input.h"

/* Example specific includes */
#include "example_common.h"
#include "nrf_mesh_config_examples.h"
#include "light_switch_example_common.h"
#include "ble_softdevice_support.h"

#define ONOFF_SERVER_0_LED          (BSP_LED_0)
#define APP_ONOFF_ELEMENT_INDEX     (0)

static bool m_device_provisioned;
uint16_t mdl_hndl;

/*************************************************************************************************/

static void node_reset(void)
{
    __LOG(LOG_SRC_APP, LOG_LEVEL_INFO, "----- Node reset  -----\n");
    hal_led_blink_ms(LEDS_MASK, LED_BLINK_INTERVAL_MS, LED_BLINK_CNT_RESET);
    /* This function may return if there are ongoing flash operations. */
    mesh_stack_device_reset();
}

static void config_server_evt_cb(const config_server_evt_t * p_evt)
{
    __LOG(LOG_SRC_APP, LOG_LEVEL_INFO, "Called: config_server_evt_cb\n")

    if (p_evt->type == CONFIG_SERVER_EVT_NODE_RESET)
    {
        node_reset();
    }
}

static void toggle_led(uint32_t pin)
{
  if(hal_led_pin_get(pin))
  {
    hal_led_pin_set(pin, false);
  }
  else
  {
    hal_led_pin_set(pin, true);
  }
}


static void button_event_handler(uint32_t button_number)
{
    __LOG(LOG_SRC_APP, LOG_LEVEL_INFO, "Button %u pressed\n", button_number);

    uint8_t data[] = {0xFF};

    uint32_t status;

    switch (button_number)
    {
        case 0:
        {
            data[0] = 0x00;
            break;
        }
        case 1:
        {
            data[0] = 0x01;
            break;
        }
        case 2:
        {
            data[0] = 0x02;
            break;
        }
        case 3:
        {
            data[0] = 0x03;
            break;
        }
    }

    const access_message_tx_t msg = {
      .opcode = ACCESS_OPCODE_VENDOR(0xC1, 0xF001),
      .p_buffer = data,
      .length = ARRAY_SIZE(data),
      .force_segmented = false,
      .transmic_size = NRF_MESH_TRANSMIC_SIZE_SMALL,
      .access_token = nrf_mesh_unique_token_get()
    };

    status = access_model_publish(mdl_hndl, &msg);
    __LOG(LOG_SRC_APP, LOG_LEVEL_INFO, "\nSend Status is %d\nModel Handle is %04X\n", status, mdl_hndl);
}

static void app_rtt_input_handler(int key)
{
    __LOG(LOG_SRC_APP, LOG_LEVEL_INFO, "RTT Received: %d \n", key)
}

static void device_identification_start_cb(uint8_t attention_duration_s)
{
    __LOG(LOG_SRC_APP, LOG_LEVEL_INFO, "Called: device_identification_start_cb\n")
    hal_led_mask_set(LEDS_MASK, false);
    hal_led_blink_ms(BSP_LED_2_MASK  | BSP_LED_3_MASK,
                     LED_BLINK_ATTENTION_INTERVAL_MS,
                     LED_BLINK_ATTENTION_COUNT(attention_duration_s));
}

static void provisioning_aborted_cb(void)
{
    hal_led_blink_stop();
}

static void provisioning_complete_cb(void)
{
    __LOG(LOG_SRC_APP, LOG_LEVEL_INFO, "Called: provisioning_complete_cb\n")
    __LOG(LOG_SRC_APP, LOG_LEVEL_INFO, "Successfully provisioned\n");

#if MESH_FEATURE_GATT_ENABLED
    /* Restores the application parameters after switching from the Provisioning
     * service to the Proxy  */
    gap_params_init();
    conn_params_init();
#endif

    dsm_local_unicast_address_t node_address;
    dsm_local_unicast_addresses_get(&node_address);
    __LOG(LOG_SRC_APP, LOG_LEVEL_INFO, "Node Address: 0x%04x \n", node_address.address_start);

    hal_led_blink_stop();
    hal_led_mask_set(LEDS_MASK, LED_MASK_STATE_OFF);
    hal_led_blink_ms(LEDS_MASK, LED_BLINK_INTERVAL_MS, LED_BLINK_CNT_PROV);
}

static void blink_leds()
{
  hal_led_blink_ms(LEDS_MASK, LED_BLINK_INTERVAL_MS, LED_BLINK_CNT_RESET);
}

static void publish_timeout_handle()
{
  
}
static void publish_interval_cb()
{

}
// Model ###########################################################################################################
static void status_handle(access_model_handle_t handle, const access_message_rx_t * p_rx_msg, void * p_args)
{
    __LOG(LOG_SRC_APP, LOG_LEVEL_INFO, "Called status_handle [Handle:%04X] \n", handle);
    __LOG(LOG_SRC_APP, LOG_LEVEL_INFO, "[Data:%02X] \n", p_rx_msg->p_data[0]);
    __LOG(LOG_SRC_APP, LOG_LEVEL_INFO, "[Source:%02X] \n", p_rx_msg->meta_data.src.value);

    toggle_led(p_rx_msg->p_data[0]);

    uint8_t data[] = {0xFF};

    const access_message_tx_t msg = {
      .opcode = ACCESS_OPCODE_VENDOR(0xC3, 0xF001),   // Build the Answer. I just noticed that i probably get the Reply-Opcode from the received Message... 
      .p_buffer = &data,
      .length = ARRAY_SIZE(data),
      .force_segmented = false,
      .transmic_size = NRF_MESH_TRANSMIC_SIZE_SMALL,
      .access_token = nrf_mesh_unique_token_get()
    };

    if(access_reliable_model_is_free())
    {
      uint32_t status = access_model_reply(handle, p_rx_msg, &msg);  // Send the Answer
      __LOG(LOG_SRC_APP, LOG_LEVEL_INFO, "Send Status is %d\n", status);
    }
}

static const access_opcode_handler_t m_opcode_handlers[] =
{
    {ACCESS_OPCODE_VENDOR(0xC2, 0xF001), status_handle},
};

static void models_init_cb(void)
{
    __LOG(LOG_SRC_APP, LOG_LEVEL_INFO, "Initializing and adding models\n");

    access_model_add_params_t add_params =
    {
        .model_id = ACCESS_MODEL_VENDOR(0xFFAB, 0xF001),            // Probably does no have an impact on the functionality
        .element_index = 0,                  // Parameter passed to this function
        .p_opcode_handlers = &m_opcode_handlers[0],      // Bind a function to an OP Code. Are OP Codes received be server?
        .opcode_count = ARRAY_SIZE(m_opcode_handlers),   // Size if opcode_handler array
        .p_args = NULL,                              // Parameter passed to this function
        .publish_timeout_cb = publish_interval_cb
    };

    uint32_t status = access_model_add(&add_params, &mdl_hndl);
    __LOG(LOG_SRC_APP, LOG_LEVEL_INFO, "\nInit Status is %d\nModel Handle is %04X\n", status, mdl_hndl);

    if (status == NRF_SUCCESS)
    {
        status = access_model_subscription_list_alloc(mdl_hndl);
        __LOG(LOG_SRC_APP, LOG_LEVEL_INFO, "\nAllocation Status is %d\nModel Handle is %04X\n", status, mdl_hndl);
    }
    
}
// #####################################################################################################################

static void mesh_init(void)
{
    __LOG(LOG_SRC_APP, LOG_LEVEL_INFO, "Called: mesh_init\n")
    mesh_stack_init_params_t init_params =
    {
        .core.irq_priority       = NRF_MESH_IRQ_PRIORITY_LOWEST,
        .core.lfclksrc           = DEV_BOARD_LF_CLK_CFG,
        .core.p_uuid             = NULL,
        .models.models_init_cb   = models_init_cb,
        .models.config_server_cb = config_server_evt_cb
    };
    ERROR_CHECK(mesh_stack_init(&init_params, &m_device_provisioned));
}

static void initialize(void)
{
    __LOG_INIT(LOG_SRC_APP | LOG_SRC_ACCESS | LOG_SRC_BEARER, LOG_LEVEL_INFO, LOG_CALLBACK_DEFAULT);
    __LOG(LOG_SRC_APP, LOG_LEVEL_INFO, "----- BLE Mesh Light Switch Server Demo -----\n");

    ERROR_CHECK(app_timer_init());
    hal_leds_init();

#if BUTTON_BOARD
    ERROR_CHECK(hal_buttons_init(button_event_handler));
#endif

    ble_stack_init();

#if MESH_FEATURE_GATT_ENABLED
    gap_params_init();
    conn_params_init();
#endif

    mesh_init();
}

static void start(void)
{
    rtt_input_enable(app_rtt_input_handler, RTT_INPUT_POLL_PERIOD_MS);

    if (!m_device_provisioned)
    {
        static const uint8_t static_auth_data[NRF_MESH_KEY_SIZE] = STATIC_AUTH_DATA;
        mesh_provisionee_start_params_t prov_start_params =
        {
            .p_static_data    = static_auth_data,
            .prov_complete_cb = provisioning_complete_cb,
            .prov_device_identification_start_cb = device_identification_start_cb,
            .prov_device_identification_stop_cb = NULL,
            .prov_abort_cb = provisioning_aborted_cb,
            .p_device_uri = EX_URI_LS_SERVER
        };
        ERROR_CHECK(mesh_provisionee_prov_start(&prov_start_params));
    }

    mesh_app_uuid_print(nrf_mesh_configure_device_uuid_get());

    ERROR_CHECK(mesh_stack_start());

    hal_led_mask_set(LEDS_MASK, LED_MASK_STATE_OFF);
    hal_led_blink_ms(LEDS_MASK, LED_BLINK_INTERVAL_MS, LED_BLINK_CNT_START);
}

int main(void)
{
    initialize();
    start();

    for (;;)
    {
        (void)sd_app_evt_wait();
    }
}
